#include "model.h"

KEY_POOL key_pool;

pthread_mutex_t key_pool_lock;

char key_charset[] = "0123456789ABCDEF";


void init_key_space(unsigned long block_size)
{
 	// init pthread_mutex
	pthread_mutex_init(&key_pool_lock, NULL);
	
	// init random seed
	double sys_time;
	#ifdef __linux__
    timeval tv;
    gettimeofday(&tv, 0);
    sys_time = tv.tv_sec + tv.tv_usec * 1e-6;
	#else
    sys_time = 1.0 * clock() / CLOCKS_PER_SEC;
	#endif
	
	srand((int)sys_time);
	
	// allocate memory for key blocks
	// do NOT generate keys at this moment, memory will drain
	key_pool.size = TOTAL_KEYS / block_size;
	key_pool.p_blocks = (KEY_BLOCK**)malloc(key_pool.size * sizeof(KEY_BLOCK*));
	key_pool.current = 0;
	unsigned long i;
	for(i = 0; i < key_pool.size; i++)
	{
		key_pool.p_blocks[i] = (KEY_BLOCK*)malloc(sizeof(KEY_BLOCK));
		key_pool.p_blocks[i]->size = block_size;
		key_pool.p_blocks[i]->start = i * block_size;
	}
	
	// randomization with Fisher-Yates shuffle
	for(i = 0; i < key_pool.size; i++)
	{
		unsigned long j = rand() * rand() % key_pool.size;
		if(j < i)
		{
			KEY_BLOCK* temp = key_pool.p_blocks[j];
			key_pool.p_blocks[j] = key_pool.p_blocks[i];
			key_pool.p_blocks[i] = temp;
		}
	}
}

void get_key_blocks(unsigned long n_req, KEY_BLOCK*** p_head, unsigned long* n_actual)
{
	// key_blocks all used up
	if(key_pool.current == key_pool.size)
	{
		*n_actual = 0;
	    return;
	}
	
	pthread_mutex_lock(&key_pool_lock);
	
	unsigned long prev_head = key_pool.current;
	
	// key_pool.current points to first unused key_block
	unsigned long blocks_left = key_pool.size - key_pool.current;
	*n_actual = blocks_left < n_req ? blocks_left : n_req;
	key_pool.current += *n_actual;
	
	// parse KEY_BLOCK* array
	KEY_BLOCK** p_blocks_ret = (KEY_BLOCK**)malloc(*n_actual * sizeof(KEY_BLOCK*));
	unsigned long i;
	for(i = 0; i < *n_actual; i++)
		p_blocks_ret[i] = key_pool.p_blocks[prev_head + i];
	*p_head = p_blocks_ret;
	
	// generate keys
	unsigned long j;
	for(i = prev_head; i < key_pool.current; i++)
	{
		KEY_BLOCK* p_block = key_pool.p_blocks[i];
		
		// memory of keys is continuous
		p_block->p_keys = (char*)malloc(p_block->size * KEY_LENGTH * sizeof(char));
		for(j = 0; j < p_block->size; j++)
		{
			unsigned long n_key = p_block->start + j;
			char* p_n_key = &n_key;
			
			int k;
			for(k = 0; k < sizeof(unsigned long); k++)
			{
				unsigned char hi = ((unsigned char)*(p_n_key + sizeof(long) - 1 - k) & 0xF0) >> 4;
				unsigned char lo = (unsigned char)*(p_n_key + sizeof(long) - 1 - k) & 0x0F;
				*(p_block->p_keys + j * KEY_LENGTH + k * 2 + 1) = key_charset[lo];
				*(p_block->p_keys + j * KEY_LENGTH + k * 2) = key_charset[hi];
			}
		}
	}
	pthread_mutex_unlock(&key_pool_lock);
}

void destroy_key_blocks(KEY_BLOCK** p_head, unsigned long size)
{
	unsigned long i;
	for(i = 0; i < size; i++)
	{
		free(p_head[i]->p_keys);
		free(p_head[i]);
	}
}

